from .rulesets import *
from .segment import *
from .general import *
import math

class controlchart: 
    # https://en.wikipedia.org/wiki/Shewhart_individuals_control_chart
    
    def __init__(self, baseline_sample_size = 8):
        self.period_series = []
        self.value_series = []
        self.segments = []
        self.mR_series = []
        self.rules = rulesets(self)
        self.baseline_sample_size = baseline_sample_size

    def load_from_pandas(self, df, period_column, value_column):
        import pandas
        self.load_from_arrays(df[period_column].tolist(), df[value_column].tolist())

    def load_from_arrays(self, period_series, value_series):
        self.mR_series = []
        self.period_series = [str(interval) for interval in period_series]
        self.value_series = value_series
        self.value_series = fillna(self.value_series, math.nan)
        self.segments = self.segment_data()
        self.mR_series = [abs(self.value_series[i+1]-self.value_series[i]) for i in range(len(self.value_series)-1)]

    # when <baseline_sample_size> samples fall the same side of the mean, recalculate the 
    # mean and standard deviations
    def segment_data(self):
        # not big enough to have more than one segment
        if (len(self.value_series) < (2 * self.baseline_sample_size)):
            pass

        # <baseline_sample_size> of 0 (or less) mean don't segment the data
        if self.baseline_sample_size < 1:
            self.baseline_sample_size = len(self.value_series)

        boundaries = []
        last_boundary = 0

        current_mean = mean(self.value_series[0:self.baseline_sample_size])
        for start in range(len(self.value_series) - self.baseline_sample_size + 1):
            # don't start a new segment close to the last
            if (start - self.baseline_sample_size) > last_boundary:
                cummulative_sum = 0
                for index in range(self.baseline_sample_size):
                    delta = self.value_series[start + index] - current_mean
                    # avoid divide by 0
                    if delta == 0:
                        side = 0
                    else:
                        side = delta / abs(delta)
                    cummulative_sum = cummulative_sum + side
                if abs(cummulative_sum) == self.baseline_sample_size:
                    boundaries.append((last_boundary, start))
                    last_boundary = start
                    current_mean = mean(self.value_series[start + 1:(start + self.baseline_sample_size)])
        # add everything else to the last segment
        boundaries.append((last_boundary, len(self.value_series)))

        # create segments out of the boundaries
        segments = []
        for boundary in boundaries:
            start, end = boundary
            segments.append(segment(self, start, end, self.baseline_sample_size))
        return segments

    # summary information
    def describe(self):
        description = 'Population' + '\n'
        description = description + '======================' + '\n'
        description = description + 'Number of Samples = ' + str(len(self.value_series)) + '\n'
        description = description + 'Population Mean = ' + '{:.3f}'.format(mean(self.value_series)) + '\n'
        description = description + 'Population Maximum = ' + str(max(self.value_series)) + '\n'
        description = description + 'Population Minimum = ' + str(min(self.value_series)) + '\n'
        description = description + 'Number of Segments = ' + str(len(self.segments)) + '\n'
        print (description)

    # collate the bin information from each of the segments
    def individuals_bins(self):
        bins = []
        for segment in self.segments:
            bins = bins + segment.individuals_bins()
        return bins

    # collate the direction information from the segments
    def individuals_direction(self):
        directions = []
        for segment in self.segments:
            directions = directions + segment.individuals_direction()
        return directions

    def individuals_sigma_line(self, number):
        sigma_line = []
        for segment in self.segments:
            sigma_line = sigma_line + segment.individuals_sigma_line(number)
        return sigma_line

    def moving_range_sigma_line(self, number):
        sigma_line = []
        for segment in self.segments:
            sigma_line = sigma_line + segment.moving_range_sigma_line(number)
        return sigma_line

    def plot(self, title="Process Behavior Chart", x_label="Period", i_label="X", mr_label="mR", file='', show_SD = False):        
        from matplotlib import pyplot as plt
        LARGE_FONT = 15
        plt.rc('font', size=LARGE_FONT)
        plt.figure(figsize=(20, 12))
        plt.title("Title")  

        # individuals
        ax = plt.subplot(211)
        plt.plot(self.period_series, self.value_series, marker='o', markersize=2, color='#6666CC')

        plt.plot(self.period_series, self.individuals_sigma_line(3), color='r', linestyle='--')
        plt.plot(self.period_series, self.individuals_sigma_line(0), color='g', linestyle='--')
        plt.plot(self.period_series, self.individuals_sigma_line(-3), color='r', linestyle='--')

        # include the SD lines on the plot
        if show_SD:
            plt.plot(self.period_series, self.individuals_sigma_line(2), color='r', linestyle=':')
            plt.plot(self.period_series, self.individuals_sigma_line(1), color='r', linestyle=':')
            plt.plot(self.period_series, self.individuals_sigma_line(-1), color='r', linestyle=':')
            plt.plot(self.period_series, self.individuals_sigma_line(-2), color='r', linestyle=':')

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        plt.title("Individuals")  
        plt.xlabel(x_label)
        plt.ylabel(i_label)
        ax.set_xticks(ax.get_xticks()[::self.baseline_sample_size])

        # moving range
        ax = plt.subplot(212)

        plt.plot(self.period_series[:-1], self.mR_series, marker='o', markersize=2, color='#6666CC')
        plt.plot(self.period_series, self.moving_range_sigma_line(3), color='r', linestyle='--')
        plt.plot(self.period_series, self.moving_range_sigma_line(0), color='g', linestyle='--')
        plt.plot(self.period_series, self.moving_range_sigma_line(-3), color='r', linestyle='--')

        # include the SD lines on the plot
        if show_SD:
            plt.plot(self.period_series, self.moving_range_sigma_line(2), color='r', linestyle=':')
            plt.plot(self.period_series, self.moving_range_sigma_line(1), color='r', linestyle=':')
            plt.plot(self.period_series, self.moving_range_sigma_line(-1), color='r', linestyle=':')
            plt.plot(self.period_series, self.moving_range_sigma_line(-2), color='r', linestyle=':')
        
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        plt.title("Moving Range")  
        plt.xlabel(x_label)
        plt.ylabel(mr_label)
        ax.set_xticks(ax.get_xticks()[::self.baseline_sample_size])

        # save the plot to a file
        if (len(file) > 0):
            plt.savefig(file, format='svg')



from .general import *

def linear_regression(X, Y): 
    x_prime = []
    y_prime = []
    
    for i in range(len(Y)):
        if (str(type(Y[i])) in num_types) and not math.isnan(Y[i]):
            x_prime.append(X[i])
            y_prime.append(Y[i])
            
    mean_x = mean(x_prime)
    mean_y = mean(y_prime)
    n = len(x_prime)
    numer = 0
    denom = 0
    for i in range(n):
        numer += (x_prime[i] - mean_x) * (y_prime[i] - mean_y)
        denom += (x_prime[i] - mean_x) ** 2.0
    m = numer / denom
    b = mean_y - (m * mean_x)

    return m, b

def mean_squares_error(x, y, m, b):
    elements = []

    for i in range(len(x)):
        if (str(type(x[i])) in num_types) and not math.isnan(x[i]) and (str(type(y[i])) in num_types) and not math.isnan(y[i]):
            lr_val = m * x[i] + b
            diff = y[i] - lr_val
            elements.append(diff * diff)
        
    return mean(elements)


from .general import *
from .linearregression import *



# split timeseries into trend, season * trend and residual
def decompose(series, period):

    if (len(series) < 10):
        raise Exception('data is too small, must be at least 10 values')

    x = list(range(len(series)))
    lr = linear_regression(x, series)

    trend = f_x(x, lambda x: x * lr[0] + lr[1])
    detrend = series_diff(series, trend)
    seasonal = (seasonal_pattern(detrend, period) * (((len(detrend) + 1) // (period - 1)) + 1))[0:len(series)]
    residual = series_diff(detrend,seasonal, 0)
    
    return decomposed_seasonal_data(series, trend, detrend, seasonal, residual)

# unweighted rolling average
def moving_average(series, window):
    import math
    roller = []
    for i in range(math.floor(window / 2) + 1):
        roller.append(math.nan)
    for i in range(len(series) - window + 1):
        roller.append (mean(series[i:i+window]))
    for i in range(math.ceil(window / 2)):
        roller.append(math.nan)
    return roller[1:len(series) + 1]

# fourier transform to identify frequencies
def series_frequencies(series):
    import numpy as np
    # to remove the DC (usually a 0 peak) subtract the mean of the set from each value
    readings= []
    series_mean = mean(series)
    for i in series:
        readings.append(i - series_mean)
    fourierTransform = np.fft.rfft(readings)
    return fourierTransform

# fourier transform to identify frequencies
def cycle_periods(series):
    fourierTransform = series_frequencies(series)
    fourierTransform = fourierTransform[0:len(fourierTransform) // 2]
    signal_mean = mean(abs(fourierTransform))
    sigma = standard_deviation(abs(fourierTransform))
    for s in range(10,0,-1):
        peaks = matches(abs(fourierTransform), lambda t: t > signal_mean + (s * sigma))
        if 1 in peaks:
            peaks.remove(1)
        if len(peaks) > 0:
            for p in range(len(peaks)):
                peaks[p] = round(len(series) / peaks[p])
            return list(set(peaks))
    raise TypeError('No cycle identified')
    
# identify cyclic pattern of series data for period
def seasonal_pattern(series, period):
    pattern = []
    cycle_sum = [0] * period
    cycle_count = [0] * period
    for i in range(len(series)):
        cycle_sum[i % period] = cycle_sum[i % period] + series[i]
        cycle_count[i % period] = cycle_count[i % period] + 1
    for i in range(period):
        cycle_mean = cycle_sum[i] / cycle_count[i]
        pattern.append(cycle_mean)
    return pattern

# apply the seasonal patterns to the trend data
def product_series(series, cycles, trend):
    if not isinstance(cycles, list): 
        cycles = [cycles]
    base = trend.copy()
    for cycle in cycles:
        pattern = seasonal_pattern(series,cycle)
        for i in range(len(series)):
            base[i] = base[i] / pattern[i % cycle]
    return base

class decomposed_seasonal_data:

    def __init__(self, source, trend, detrend, seasonal, residual):
        self.source = source
        self.trend = trend
        self.detrend = detrend
        self.seasonal = seasonal
        self.residual = residual
        self.record = len(self.source)

    def plot(self):
        from matplotlib import pyplot as plt
        LARGE_FONT = 15
        plt.rc('font', size=LARGE_FONT)
        plt.figure(figsize=(20, 12))
        ax = plt.subplot(111)

        # source
        ax = plt.subplot(511)
        plt.ylabel('Source')
        ax.axhline(y=mean(self.source), color='#AAAAAA', linestyle=':')
        plt.plot(self.source, color='#6666CC')

        # regression
        ax = plt.subplot(512)
        plt.ylabel('Regression')
        ax.axhline(y=mean(self.trend), color='#AAAAAA', linestyle=':')
        plt.plot(self.trend, color='#6666CC')

        # seasonal pattern
        ax = plt.subplot(513)
        plt.ylabel('Seasonal')
        ax.axhline(y=mean(self.seasonal), color='#AAAAAA', linestyle=':')
        plt.plot(self.seasonal, color='#6666CC')
        
        # detrend
        ax = plt.subplot(514)
        plt.ylabel('Detrend')
        ax.axhline(y=mean(self.detrend), color='#AAAAAA', linestyle=':')
        plt.plot(self.detrend, color='#6666CC')

        # residual 
        ax = plt.subplot(515)
        plt.ylabel('Residual')
        ax.axhline(y=mean(self.residual), color='#AAAAAA', linestyle=':')
        plt.plot(self.residual, color='#6666CC')

        from matplotlib import pyplot as plt


class rulesets:

    mr = None

    def __init__(self, mr):
        self.mr = mr

    # https://en.wikipedia.org/wiki/Nelson_rules
    def nelson(self):
        violations = { }
        bins = self.mr.individuals_bins()

        # Rule 1
        # One point is more than 3σ from the mean.
        # One sample is out of control.
        rule1 = execute_run_test(bins, 1, lambda t: t > 3 or t < -3)
        violations['Points more than 3σ from the mean'] = rule1  

        # Rule 2
        # Nine (or more) points in a row are on the same side of the mean.
        # Some prolonged bias exists.
        rule2 = execute_run_test(bins, 9, lambda t: t > 0)
        rule2 = rule2 + execute_run_test(bins, 9, lambda t: t < 0)
        violations['9 consecutive points same side of the mean'] = sorted(rule2)

        # Rule 3
        # Six (or more) points in a row are continually increasing (or decreasing).
        # A trend exists.
        direction = self.mr.individuals_direction()
        rule3 = execute_run_test(direction, 6, lambda t: t > 0)
        rule3 = rule3 + execute_run_test(direction, 6, lambda t: t < 0)
        violations['6 consecutive increase or decreasing points'] = sorted(rule3)

        # Rule 4
        # Fourteen (or more) points in a row alternate in direction, increasing then decreasing.
        # This much oscillation is beyond noise. Note that the rule is concerned with directionality only. The position of the mean and the size of the standard deviation have no bearing.
        #
        # This is the only rule, not able to be implemented standard method
        rule4 = []
        rolling = [0] * 14
        for sample in range(len(direction) - 1):
            rolling.append(direction[sample] != direction[sample + 1])
            rolling = rolling[1:]
            if sum(rolling) == 14:
                for index in range(14):
                    rule4.append(sample - index)
        violations['14 consecutive oscillating points'] = sorted(list(set(rule4)))
        
        # Rule 5
        # Two (or three) out of three points in a row are more than 2σ from the mean in the same direction.
        # There is a medium tendency for samples to be mediumly out of control.The position of the third point is unspecified.
        rule5 = execute_run_test(bins, 3, lambda t: t > 2, 2/3)
        rule5 = rule5 + execute_run_test(bins, 3, lambda t: t < -2, 2/3)
        violations['2 or more of 3 consecutive points more than 2σ from the mean'] = sorted(rule5)

        # Rule 6
        # Four (or five) out of five points in a row are more than 1σ from the mean in the same direction.
        # There is a strong tendency for samples to be slightly out of control. The position of the fifth point is unspecified.
        rule6 = execute_run_test(bins, 5, lambda t: t > 1, 4/5)
        rule6 = rule6 + execute_run_test(bins, 5, lambda t: t < -1, 4/5)
        violations['4 or more of 5 consecutive points more than 1σ from the mean'] = sorted(rule6)

        # Rule 7
        # Fifteen points in a row are all within 1σ of the mean on either side of the mean.
        # With 1σ, greater variation would be expected.
        rule7 = execute_run_test(bins, 15, lambda t: abs(t) == 1)
        violations['15 consecutive points within 1σ of the mean'] = sorted(rule7)

        # Rule 8
        # Eight points in a row exist, but none within 1σ of the mean, and the points are in both directions from the mean.
        # Jumping from above to below whilst missing the 1σ band is rarely random.
        rule8 = execute_run_test(bins, 8, lambda t: abs(t) > 1)
        violations['8 consecutive points outwith 1σ of the mean'] = sorted(rule8)

        return violations
        
    # https://en.wikipedia.org/wiki/Western_Electric_rules
    def western_electic(self):
        violations = { }
        bins = self.mr.individuals_bins()

        # Rule 1
        # Any single data point falls outside the 3σ-limit from the mean
        rule1 = execute_run_test(bins, 1, lambda t: t > 3 or t < -3)
        violations['Points more than 3σ from the mean'] = rule1        

        # Rule 2
        # Two out of three consecutive points fall beyond the 2σ-limit, on the same side of the mean
        rule2 = execute_run_test(bins, 3, lambda t: t > 2, 2/3)
        rule2 = rule2 + execute_run_test(bins, 3, lambda t: t < -2, 2/3)
        violations['2 or more of 3 consecutive points more than 2σ from the mean'] = sorted(rule2)

        # Rule 3
        # Four out of five consecutive points fall beyond the 1σ-limit, on the same side of the mean
        rule3 = execute_run_test(bins, 5, lambda t: t > 1, 4/5)
        rule3 = rule3 + execute_run_test(bins, 5, lambda t: t < -1, 4/5)
        violations['4 or more of 5 consecutive points more than 1σ from the mean'] = sorted(rule3)

        # Rule 4
        # Eight consecutive points fall on the same side of the mean
        rule4 = execute_run_test(bins, 8, lambda t: t > 0)
        rule4 = rule4 + execute_run_test(bins, 8, lambda t: t < 0)
        violations['8 consecutive points same side of the mean'] = sorted(rule4)

        return violations

    def basic(self):
        violations = { }
        bins = self.mr.individuals_bins()

        # Rule 1
        # Any single data point falls outside the 3σ-limit from the mean
        rule1 = execute_run_test(bins, 1, lambda t: t > 3 or t < -3)
        violations['Points more than 3σ from the mean'] = rule1

        # Rule 2
        # Seven consecutive points fall on the same side of the mean
        rule2 = execute_run_test(bins, 7, lambda t: t > 0)
        rule2 = rule2 + execute_run_test(bins, 7, lambda t: t < 0)
        violations['7 consecutive points same side of the mean'] = sorted(rule2)

        return violations

# test should result in a boolean
# method tests runs of length sample size return true for the test some rules are 
# "four out of five", an optional ratio parameter allows for this
def execute_run_test(samples, sample_size, test, ratio = 1):
    results = []
    rolling = [0] * sample_size
    for sample in range(len(samples)):
        rolling.append(test(samples[sample]))
        rolling = rolling[1:]
        if sum(rolling) >= (sample_size * ratio):
            for index in range(sample_size):
                results.append(sample - index)
    return sorted(list(set(results)))


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
   name='va-timeseries',
   version='1.0.7',
   description='Timeseries Analysis',
   long_description=long_description,
   long_description_content_type="text/markdown",
   author='Joocer',
   author_email='justin.joyce@joocer.com',
   packages=find_packages(),
   url="https://github.com/joocer/timeseries",
   install_requires=['pandas', 'matplotlib', 'numpy']
)